/**
 * From the OpenGL Programming wikibook: http://en.wikibooks.org/wiki/OpenGL_Programming
 * This file is in the public domain.
 * Contributors: Martin Kraus, Sylvain Beucler
 */

in vec4 position;  // position of the vertex (and fragment) in world space
in vec3 varyingNormalDirection;  // surface normal vector in world space
in vec2 f_texcoord;
in vec4 ShadowCoord;
in vec3 LightDirection_cameraspace;
in vec3 normal;
in float z_linear;
in float depth;
in vec3 ViewRay;

uniform mat4 m, v, p;
uniform mat4 v_inv;
uniform vec3 lightPos; // light position in world space
uniform sampler2D objTexture;
uniform sampler2D ShadowMap;

out vec4 frag_color;
out vec4 frag_normal;
out vec4 frag_viewspaceDepth;
// out vec4 frag_viewray;
 

// Poisson Disk for shadow
vec2 poissonDisk[4] = vec2[](
  vec2( -0.94201624, -0.39906216 ),
  vec2( 0.94558609, -0.76890725 ),
  vec2( -0.094184101, -0.92938870 ),
  vec2( 0.34495938, 0.29387760 )
);

struct lightSource
{
  vec4 position;
  vec4 diffuse;		 // Red, Green, Blue, Alpha
  vec4 specular;
  float constantAttenuation, linearAttenuation, quadraticAttenuation;
  float spotCutoff, spotExponent;
  vec3 spotDirection;
};
const int numberOfLights = 1;
lightSource lights[numberOfLights];
lightSource light0 = lightSource(
  vec4(lightPos, 1.0),
  vec4(0.95, 0.95, 0.95, 1.0),
  vec4(0.5,  0.5,  0.5, 1.0), // specula
  0.0, 0.1, 0.0,
  180.0, 0.0,
  vec3(0.0, 0.0, 0.0)
);


vec4 scene_ambient = vec4(0.5, 0.5, 0.5, 1.0);

struct material
{
  vec4 ambient;
  vec4 diffuse;
  vec4 specular;
  float shininess;
};
material frontMaterial = material(
  vec4(1.0, 1.0, 1.0, 1.0),
  vec4(1.0, 1.0, 1.0, 1.0),
  vec4(1.0, 1.0, 1.0, 1.0),
  5.0
);

void main()
{
 // ----------------------------------------------------
 // Here starts the calculation for phong component
 // ----------------------------------------------------
  lights[0] = light0;
 // lights[1] = light1;

  vec3 normalDirection = normalize(varyingNormalDirection);
  vec3 viewDirection = normalize(vec3(v_inv * vec4(0.0, 0.0, 0.0, 1.0) - position));
  vec3 lightDirection;
  float attenuation;
  
  // initialize total lighting with ambient lighting
  vec3 totalLighting = vec3(scene_ambient) * vec3(frontMaterial.ambient);
  
  for (int index = 0; index < numberOfLights; index++) // for all light sources
  {
      if (0.0 == lights[index].position.w) // directional light?
	{
	  attenuation = 1.0; // no attenuation
	  lightDirection = normalize(vec3(lights[index].position));
	} 
      else // point light or spotlight (or other kind of light) 
	{
	  vec3 positionToLightSource = vec3(lights[index].position - position);
	  float distance = length(positionToLightSource);
	  lightDirection = normalize(positionToLightSource);
	  attenuation = 1.0 / (lights[index].constantAttenuation
			       + lights[index].linearAttenuation * distance
			       + lights[index].quadraticAttenuation * distance * distance);
	  
	  if (lights[index].spotCutoff <= 90.0) // spotlight?
	    {
	      float clampedCosine = max(0.0, dot(-lightDirection, normalize(lights[index].spotDirection)));
	      if (clampedCosine < cos(radians(lights[index].spotCutoff))) // outside of spotlight cone?
		{
		  attenuation = 0.0;
		}
	      else
		{
		  attenuation = attenuation * pow(clampedCosine, lights[index].spotExponent);   
		}
	    }
	}
      
      vec3 diffuseReflection = attenuation 
	* vec3(lights[index].diffuse) * vec3(frontMaterial.diffuse)
	* max(0.0, dot(normalDirection, lightDirection));
      
      vec3 specularReflection;
      if (dot(normalDirection, lightDirection) < 0.0) // light source on the wrong side?
	{
	  specularReflection = vec3(0.0, 0.0, 0.0); // no specular reflection
	}
      else // light source on the right side
	{
	  specularReflection = attenuation * vec3(lights[index].specular) * vec3(frontMaterial.specular)
	    * pow(max(0.0, dot(reflect(-lightDirection, normalDirection), viewDirection)), frontMaterial.shininess);
	}

      totalLighting = totalLighting + diffuseReflection + specularReflection;
    }
  
    vec4 phong_components = vec4(totalLighting, 1.0);
    vec4 tex_color = texture2D(objTexture, f_texcoord);

	// -----------------------------------------
	// Here starts the calculation for shadow
	// -----------------------------------------
	vec3 Normal_cameraspace = ( v * vec4(varyingNormalDirection,0)).xyz; 
	vec3 n = normalize( Normal_cameraspace ); // Normal of the computed fragment, in camera space
	vec3 l = normalize( LightDirection_cameraspace ); // Direction of the light (from the fragment to the light)
	// Cosine of the angle between the normal and the light direction, 
	// clamped above 0
	//  - light is at the vertical of the triangle -> 1
	//  - light is perpendiular to the triangle -> 0
	//  - light is behind the triangle -> 0
	float cosTheta = clamp( dot( n,l ), 0,1 );

	float bias = 0.02*tan(acos(cosTheta));
	bias = clamp(bias, 0,0.03);

	float visibility = 1.0;

	for (int i=0;i<4;i++){
		if ( texture2D( ShadowMap, ShadowCoord.xy/ShadowCoord.w + poissonDisk[i]/1000.0 ).z  <  (ShadowCoord.z-bias)/ShadowCoord.w ){
		   visibility-=0.15;
		}
    }


    frag_color =  vec4(visibility *  phong_components.x * tex_color.x, visibility * phong_components.y * tex_color.y, visibility * phong_components.z * tex_color.z, 1.0 );

    vec3 normalVec = normalize(normal);
	//frag_normal = vec4(normalVec,  -z_linear);
	frag_normal = vec4((normalVec + vec3(1.0))/2.0,  1.0);
	frag_viewspaceDepth = vec4(-z_linear,  -z_linear, -z_linear, 1.0);
	
	//frag_viewray = vec4(ViewRay.x, ViewRay.y, -ViewRay.z,1.0);
}
